from scipy.optimize import fsolve, curve_fit
from numpy import log as ln
from scipy.integrate import odeint, quad
import numpy as np
from numpy import sin,cos,tan
from numpy import sinh, cosh,tanh
import matplotlib.pyplot as plt
from scipy.interpolate import interp1d
import os
import pandas as pd